from abc import ABC

from dygie.models.shared import fields_to_batches, batches_to_fields
import copy
import numpy as np
import re
import json


def format_float(x):
    return round(x, 4)


class SpanCrossesSentencesError(ValueError):
    pass


def get_sentence_of_span(span, sentence_starts, doc_tokens):
    """
    Return the index of the sentence that the span is part of.
    """
    # Inclusive sentence ends
    sentence_ends = [x - 1 for x in sentence_starts[1:]] + [doc_tokens - 1]
    in_between = [span[0] >= start and span[1] <= end
                  for start, end in zip(sentence_starts, sentence_ends)]
    if sum(in_between) != 1:
        raise SpanCrossesSentencesError
    the_sentence = in_between.index(True)
    return the_sentence


def update_sentences_with_clusters(sentences, clusters):
    "Add cluster dictionary to each sentence, if there are coreference clusters."
    for sent in sentences:
        sent.cluster_dict = {} if clusters is not None else None

    if clusters is None:
        return sentences

    for clust in clusters:
        for member in clust.members:
            sent = member.sentence
            sent.cluster_dict[member.span.span_sent] = member.cluster_id

    return sentences


def update_sentences_with_event_clusters(sentences, event_clusters):
    "Add event cluster dictionary to each sentence, if there are event coreference clusters."
    for sent in sentences:
        sent.event_cluster_dict = {} if event_clusters is not None else None

    if event_clusters is None:
        return sentences

    for event_clust in event_clusters:
        for member in event_clust.members:
            sent = member.sentence
            sent.event_cluster_dict[member.span.span_sent] = member.cluster_id

    return sentences


class Dataset:
    def __init__(self, documents):
        self.documents = documents

    def __getitem__(self, i):
        return self.documents[i]

    def __len__(self):
        return len(self.documents)

    def __repr__(self):
        return f"Dataset with {self.__len__()} documents."

    @classmethod
    def from_jsonl(cls, fname):
        documents = []
        with open(fname, "r") as f:
            for line in f:
                doc = Document.from_json(json.loads(line))
                documents.append(doc)

        return cls(documents)

    def to_jsonl(self, fname):
        to_write = [doc.to_json() for doc in self]
        with open(fname, "w") as f:
            for entry in to_write:
                print(json.dumps(entry), file=f)


class Document:
    def __init__(self, doc_key, dataset, sentences,
                 clusters=None, predicted_clusters=None, event_clusters=None, predicted_event_clusters=None, weight=None):
        self.doc_key = doc_key
        self.dataset = dataset
        self.sentences = sentences
        self.clusters = clusters
        self.predicted_clusters = predicted_clusters
        self.event_clusters = event_clusters
        self.predicted_event_clusters = predicted_event_clusters
        self.weight = weight

    @classmethod
    def from_json(cls, js):
        "Read in from json-loaded dict."
        cls._check_fields(js)
        doc_key = js["doc_key"]
        dataset = js.get("dataset")
        entries = fields_to_batches(js, ["doc_key", "dataset", "clusters", "predicted_clusters",
                                         "weight", "event_clusters","predicted_event_clusters"])
        sentence_lengths = [len(entry["sentences"]) for entry in entries]
        sentence_starts = np.cumsum(sentence_lengths)
        sentence_starts = np.roll(sentence_starts, 1)
        sentence_starts[0] = 0
        sentence_starts = sentence_starts.tolist()
        sentences = [Sentence(entry, sentence_start, sentence_ix)
                     for sentence_ix, (entry, sentence_start)
                     in enumerate(zip(entries, sentence_starts))]
        # Store cofereference annotations.
        if "clusters" in js:
            clusters = [Cluster(entry, i, sentences, sentence_starts)
                        for i, entry in enumerate(js["clusters"])]
        else:
            clusters = None
        # TODO(dwadden) Need to treat predicted clusters differently and update sentences
        # appropriately.

        if "predicted_clusters" in js:
            predicted_clusters = [Cluster(entry, i, sentences, sentence_starts)
                                  for i, entry in enumerate(js["predicted_clusters"])]
        else:
            predicted_clusters = None

        # adapt from entity clusters
        if "event_clusters" in js:
            event_clusters = [Cluster(entry, i, sentences, sentence_starts)
                              for i, entry in enumerate(js["event_clusters"])]
        else:
            event_clusters = None
        # TODO(dwadden) Need to treat predicted clusters differently and update sentences
        # appropriately.
        if "predicted_event_clusters" in js:
            predicted_event_clusters = [Cluster(entry, i, sentences, sentence_starts)
                                        for i, entry in enumerate(js["predicted_event_clusters"])]
        else:
            predicted_event_clusters = None

        # Update the sentences with coreference cluster labels.
        sentences = update_sentences_with_clusters(sentences, clusters)
        sentences = update_sentences_with_event_clusters(sentences, event_clusters)
        # Get the loss weight for this document.
        weight = js.get("weight", None)

        return cls(doc_key, dataset, sentences, clusters, predicted_clusters, event_clusters,
                   predicted_event_clusters, weight)

    @staticmethod
    def _check_fields(js):
        "Make sure we only have allowed fields."
        allowed_field_regex = ("doc_key|dataset|sentences|weight|.*ner$|"
                               ".*relations$|.*clusters$|.*events$|^_.*")
        allowed_field_regex = re.compile(allowed_field_regex)
        unexpected = []
        for field in js.keys():
            if not allowed_field_regex.match(field):
                unexpected.append(field)

        if unexpected:
            msg = f"The following unexpected fields should be prefixed with an underscore: {', '.join(unexpected)}."
            raise ValueError(msg)

    def to_json(self):
        "Write to json dict."
        res = {"doc_key": self.doc_key,
               "dataset": self.dataset}
        sents_json = [sent.to_json() for sent in self]
        fields_json = batches_to_fields(sents_json)
        res.update(fields_json)
        if self.clusters is not None:
            res["clusters"] = [cluster.to_json() for cluster in self.clusters]
        if self.predicted_clusters is not None:
            res["predicted_clusters"] = [cluster.to_json() for cluster in self.predicted_clusters]
        if self.event_clusters is not None:
            res["event_clusters"] = [cluster.to_json() for cluster in self.event_clusters]
        if self.predicted_event_clusters is not None:
            res["predicted_event_clusters"] = [cluster.to_json() for cluster in self.predicted_event_clusters]

        if self.weight is not None:
            res["weight"] = self.weight

        return res

    # TODO(dwadden) Write a unit test to make sure this does the correct thing.
    def split(self, max_tokens_per_doc):
        """
        Greedily split a long document into smaller documents, each shorter than
        `max_tokens_per_doc`. Each split document will get the same weight as its parent.
        """
        # TODO(dwadden) Implement splitting when there's coref annotations. This is more difficult
        # because coreference clusters have to be split across documents.
        if self.clusters is not None or self.predicted_clusters is not None:
            raise NotImplementedError("Splitting documents with coreference annotations not implemented.")
        if self.event_clusters is not None or self.predicted_event_clusters is not None:
            raise NotImplementedError("Splitting documents with event coreference annotations not implemented.")

        # If the document is already short enough, return it as a list with a single item.
        if self.n_tokens <= max_tokens_per_doc:
            return [self]

        sentences = copy.deepcopy(self.sentences)

        sentence_groups = []
        current_group = []
        group_length = 0
        sentence_tok_offset = 0
        sentence_ix_offset = 0
        for sentence in sentences:
            # Can't deal with single sentences longer than the limit.
            if len(sentence) > max_tokens_per_doc:
                msg = f"Sentence \"{''.join(sentence.text)}\" has more than {max_tokens_per_doc} tokens. Please split this sentence."
                raise ValueError(msg)

            if group_length + len(sentence) <= max_tokens_per_doc:
                # If we're not at the limit, add it to the current sentence group.
                sentence.sentence_start -= sentence_tok_offset
                sentence.sentence_ix -= sentence_ix_offset
                current_group.append(sentence)
                group_length += len(sentence)
            else:
                # Otherwise, start a new sentence group and adjust sentence offsets.
                sentence_groups.append(current_group)
                sentence_tok_offset = sentence.sentence_start
                sentence_ix_offset = sentence.sentence_ix
                sentence.sentence_start -= sentence_tok_offset
                sentence.sentence_ix -= sentence_ix_offset
                current_group = [sentence]
                group_length = len(sentence)

        # Add the final sentence group.
        sentence_groups.append(current_group)

        # Create a separate document for each sentence group.
        doc_keys = [f"{self.doc_key}_SPLIT_{i}" for i in range(len(sentence_groups))]
        res = [self.__class__(doc_key, self.dataset, sentence_group,
                              self.clusters, self.predicted_clusters, self.weight)
               for doc_key, sentence_group in zip(doc_keys, sentence_groups)]

        return res

    def __repr__(self):
        return "\n".join([str(i) + ": " + " ".join(sent.text) for i, sent in enumerate(self.sentences)])

    def __getitem__(self, ix):
        return self.sentences[ix]

    def __len__(self):
        return len(self.sentences)

    def print_plaintext(self):
        for sent in self:
            print(" ".join(sent.text))

    @property
    def n_tokens(self):
        return sum([len(sent) for sent in self.sentences])

    def find_cluster(self, entity):
        """
        Search through coreference clusters and return the one containing the query entity, if it's
        part of a cluster. If we don't find a match, return None.
        """
        for clust in self.clusters:
            for entry in clust:
                if entry.span == entity.span:
                    return clust

        return None

    @property
    def n_tokens(self):
        return sum([len(sent) for sent in self.sentences])


class Sentence:
    def __init__(self, entry, sentence_start, sentence_ix):
        self.sentence_start = sentence_start
        self.sentence_ix = sentence_ix
        self.text = entry["sentences"]

        # Metadata fields are prefixed with a `_`.
        self.metadata = {k: v for k, v in entry.items() if re.match("^_", k)}

        # Store events.
        if "ner" in entry:
            self.ner = [NER(this_ner, self)
                        for this_ner in entry["ner"]]
            self.ner_dict = {entry.span.span_sent: entry.label for entry in self.ner}
        else:
            self.ner = None
            self.ner_dict = None

        # Predicted ner.
        if "predicted_ner" in entry:
            self.predicted_ner = [PredictedNER(this_ner, self)
                                  for this_ner in entry["predicted_ner"]]
        else:
            self.predicted_ner = None

        # Store relations.
        if "relations" in entry:
            self.relations = [Relation(this_relation, self) for
                              this_relation in entry["relations"]]
            relation_dict = {}
            for rel in self.relations:
                key = (rel.pair[0].span_sent, rel.pair[1].span_sent)
                relation_dict[key] = rel.label
            self.relation_dict = relation_dict
        else:
            self.relations = None
            self.relation_dict = None

        # Predicted relations.
        if "predicted_relations" in entry:
            self.predicted_relations = [PredictedRelation(this_relation, self) for
                                        this_relation in entry["predicted_relations"]]
        else:
            self.predicted_relations = None

        # Store events.
        if "events" in entry:
            self.events = Events(entry["events"], self)
        else:
            self.events = None

        # Predicted events.
        if "predicted_events" in entry:
            self.predicted_events = PredictedEvents(entry["predicted_events"], self)
        else:
            self.predicted_events = None

    def to_json(self):
        res = {"sentences": self.text}
        if self.ner is not None:
            res["ner"] = [entry.to_json() for entry in self.ner]
        if self.predicted_ner is not None:
            res["predicted_ner"] = [entry.to_json() for entry in self.predicted_ner]
        if self.relations is not None:
            res["relations"] = [entry.to_json() for entry in self.relations]
        if self.predicted_relations is not None:
            res["predicted_relations"] = [entry.to_json() for entry in self.predicted_relations]
        if self.events is not None:
            res["events"] = self.events.to_json()
        if self.predicted_events is not None:
            res["predicted_events"] = self.predicted_events.to_json()

        for k, v in self.metadata.items():
            res[k] = v

        return res

    def __repr__(self):
        the_text = " ".join(self.text)
        the_lengths = [len(x) for x in self.text]
        tok_ixs = ""
        for i, offset in enumerate(the_lengths):
            true_offset = offset if i < 10 else offset - 1
            tok_ixs += str(i)
            tok_ixs += " " * true_offset

        return the_text + "\n" + tok_ixs

    def __len__(self):
        return len(self.text)


class Span:
    def __init__(self, start, end, sentence, sentence_offsets=False):
        # The `start` and `end` are relative to the document. We convert them to be relative to the
        # sentence.
        self.sentence = sentence
        # Need to store the sentence text to make span objects hashable.
        self.sentence_text = " ".join(sentence.text)
        self.start_sent = start if sentence_offsets else start - sentence.sentence_start
        self.end_sent = end if sentence_offsets else end - sentence.sentence_start

    @property
    def start_doc(self):
        return self.start_sent + self.sentence.sentence_start

    @property
    def end_doc(self):
        return self.end_sent + self.sentence.sentence_start

    @property
    def span_doc(self):
        return (self.start_doc, self.end_doc)

    @property
    def span_sent(self):
        return (self.start_sent, self.end_sent)

    @property
    def text(self):
        return self.sentence.text[self.start_sent:self.end_sent + 1]

    def __repr__(self):
        return str((self.start_sent, self.end_sent, self.text))

    def __eq__(self, other):
        return (self.span_doc == other.span_doc and
                self.span_sent == other.span_sent and
                self.sentence == other.sentence)

    def __hash__(self):
        tup = self.span_sent + (self.sentence_text,)
        return hash(tup)


class Token:
    def __init__(self, ix, sentence, sentence_offsets=False):
        self.sentence = sentence
        self.ix_sent = ix if sentence_offsets else ix - sentence.sentence_start

    @property
    def ix_doc(self):
        return self.ix_sent + self.sentence.sentence_start

    @property
    def text(self):
        return self.sentence.text[self.ix_sent]

    def __repr__(self):
        return str((self.ix_sent, self.text))


class Trigger:
    def __init__(self, trig, sentence, sentence_offsets):
        token = Token(trig[0], sentence, sentence_offsets)
        label = trig[1]
        self.token = token
        self.label = label

    def __repr__(self):
        return self.token.__repr__()[:-1] + ", " + self.label + ")"

    def to_json(self):
        return [self.token.ix_doc, self.label]


class PredictedTrigger(Trigger):
    def __init__(self, trig, sentence, sentence_offsets):
        super().__init__(trig, sentence, sentence_offsets)
        self.raw_score = trig[2]
        self.softmax_score = trig[3]

    def __repr__(self):
        return super().__repr__() + f" with confidence {self.softmax_score:0.4f}"

    def to_json(self):
        return super().to_json() + [format_float(self.raw_score), format_float(self.softmax_score)]


class Argument:
    def __init__(self, arg, event_type, sentence, sentence_offsets):
        self.span = Span(arg[0], arg[1], sentence, sentence_offsets)
        self.role = arg[2]
        self.event_type = event_type

    def __repr__(self):
        return self.span.__repr__()[:-1] + ", " + self.event_type + ", " + self.role + ")"

    def __eq__(self, other):
        return (self.span == other.span and
                self.role == other.role and
                self.event_type == other.event_type)

    def __hash__(self):
        return self.span.__hash__() + hash((self.role, self.event_type))

    def to_json(self):
        return list(self.span.span_doc) + [self.role]


class PredictedArgument(Argument):
    def __init__(self, arg, event_type, sentence, sentence_offsets):
        super().__init__(arg, event_type, sentence, sentence_offsets)
        self.raw_score = arg[3]
        self.softmax_score = arg[4]

    def __repr__(self):
        return super().__repr__() + f" with confidence {self.softmax_score:0.4f}"

    def to_json(self):
        return super().to_json() + [format_float(self.raw_score), format_float(self.softmax_score)]


class NER:
    def __init__(self, ner, sentence, sentence_offsets=False):
        self.span = Span(ner[0], ner[1], sentence, sentence_offsets)
        self.label = ner[2]

    def __repr__(self):
        return f"{self.span.__repr__()}: {self.label}"

    def __eq__(self, other):
        return (self.span == other.span and
                self.label == other.label)

    def to_json(self):
        return list(self.span.span_doc) + [self.label]


class PredictedNER(NER):
    def __init__(self, ner, sentence, sentence_offsets=False):
        "The input should be a list: [span_start, span_end, label, raw_score, softmax_score]."
        super().__init__(ner, sentence, sentence_offsets)
        self.raw_score = ner[3]
        self.softmax_score = ner[4]

    def __repr__(self):
        return super().__repr__() + f" with confidence {self.softmax_score:0.4f}"

    def to_json(self):
        return super().to_json() + [format_float(self.raw_score), format_float(self.softmax_score)]


class Relation:
    def __init__(self, relation, sentence, sentence_offsets=False):
        start1, end1 = relation[0], relation[1]
        start2, end2 = relation[2], relation[3]
        label = relation[4]
        span1 = Span(start1, end1, sentence, sentence_offsets)
        span2 = Span(start2, end2, sentence, sentence_offsets)
        self.pair = (span1, span2)
        self.label = label

    def __repr__(self):
        return f"{self.pair[0].__repr__()}, {self.pair[1].__repr__()}: {self.label}"

    def __eq__(self, other):
        return (self.pair == other.pair) and (self.label == other.label)

    def to_json(self):
        return list(self.pair[0].span_doc) + list(self.pair[1].span_doc) + [self.label]


class PredictedRelation(Relation):
    def __init__(self, relation, sentence, sentence_offsets=False):
        "Input format: [start_1, end_1, start_2, end_2, label, raw_score, softmax_score]."
        super().__init__(relation, sentence, sentence_offsets)
        self.raw_score = relation[5]
        self.softmax_score = relation[6]

    def __repr__(self):
        return super().__repr__() + f" with confidence {self.softmax_score:0.4f}"

    def to_json(self):
        return super().to_json() + [format_float(self.raw_score), format_float(self.softmax_score)]


# This code is a little tricky. We want Events to use Triggers and Arguments, while PredictedEvents
# use PredictedTriggers and PredictedArguments. I create a base class that defines the methods, and
# then subclasses set the constructors to be used.
class EventBase(ABC):
    def __init__(self, event, sentence, sentence_offsets=False):
        trig = event[0]
        args = event[1:]
        self.trigger = self.trigger_constructor(trig, sentence, sentence_offsets)

        self.arguments = []
        for arg in args:
            this_arg = self.argument_constructor(arg, self.trigger.label, sentence, sentence_offsets)
            self.arguments.append(this_arg)

    def to_json(self):
        trig_json = self.trigger.to_json()
        arg_json = [arg.to_json() for arg in self.arguments]
        res = [trig_json] + arg_json
        return res

    def __repr__(self):
        res = "<"
        res += self.trigger.__repr__() + ":\n"
        for arg in self.arguments:
            res += 6 * " " + arg.__repr__() + ";\n"
        res = res[:-2] + ">"
        return res


class Event(EventBase):
    trigger_constructor = Trigger
    argument_constructor = Argument


class PredictedEvent(EventBase):
    trigger_constructor = PredictedTrigger
    argument_constructor = PredictedArgument


# Same pattern as above. Define base class, and pass constructors to child classes.
class EventsBase(ABC):
    def __init__(self, events_json, sentence, sentence_offsets=False):
        self.event_list = [self.event_constructor(this_event, sentence, sentence_offsets)
                           for this_event in events_json]
        self.triggers = set([event.trigger for event in self.event_list])
        self.arguments = set([arg for event in self.event_list for arg in event.arguments])

        # Store trigger and argument dictionaries.
        trigger_dict = {}
        argument_dict = {}
        for event in self.event_list:
            trigger_key = event.trigger.token.ix_sent  # integer index
            trigger_val = event.trigger.label          # trigger label
            trigger_dict[trigger_key] = trigger_val
            for argument in event.arguments:
                arg_key = (trigger_key, argument.span.span_sent)  # (trigger_ix, (arg_start, arg_end))
                arg_value = argument.role                         # argument label
                argument_dict[arg_key] = arg_value

        self.trigger_dict = trigger_dict
        self.argument_dict = argument_dict

    def to_json(self):
        return [event.to_json() for event in self]

    def __len__(self):
        return len(self.event_list)

    def __getitem__(self, i):
        return self.event_list[i]

    def __repr__(self):
        return "\n\n".join([event.__repr__() for event in self.event_list])

    def span_matches(self, argument):
        return set([candidate for candidate in self.arguments
                    if candidate.span.span_sent == argument.span.span_sent])

    def event_type_matches(self, argument):
        return set([candidate for candidate in self.span_matches(argument)
                    if candidate.event_type == argument.event_type])

    def matches_except_event_type(self, argument):
        matched = [candidate for candidate in self.span_matches(argument)
                   if candidate.event_type != argument.event_type
                   and candidate.role == argument.role]
        return set(matched)

    def exact_match(self, argument):
        for candidate in self.arguments:
            if candidate == argument:
                return True
        return False


class Events(EventsBase):
    event_constructor = Event


class PredictedEvents(EventsBase):
    event_constructor = PredictedEvent


class Cluster:
    def __init__(self, cluster, cluster_id, sentences, sentence_starts):
        # Make sure the cluster ID is an int.
        if not isinstance(cluster_id, int):
            raise TypeError("Coreference cluster ID's must be ints.")

        n_tokens = sum([len(x) for x in sentences])

        members = []
        members_crossing_sentences = []

        for entry in cluster:
            try:
                sentence_ix = get_sentence_of_span(entry, sentence_starts, n_tokens)
                sentence = sentences[sentence_ix]
                span = Span(entry[0], entry[1], sentence)
                to_append = ClusterMember(span, sentence, cluster_id)
                members.append(to_append)
            except SpanCrossesSentencesError:
                members_crossing_sentences.append(entry)

        if members_crossing_sentences:
            print("Found a coreference cluster member that crosses sentence boundaries; skipping.")

        self.members = members
        self.cluster_id = cluster_id

    def to_json(self):
        return [list(member.span.span_doc) for member in self.members]

    def __repr__(self):
        return f"{self.cluster_id}: " + self.members.__repr__()

    def __getitem__(self, ix):
        return self.members[ix]

    def __len__(self):
        return len(self.members)


class ClusterMember:
    def __init__(self, span, sentence, cluster_id):
        self.span = span
        self.sentence = sentence
        self.cluster_id = cluster_id

    def __repr__(self):
        return f"<{self.sentence.sentence_ix}> " + self.span.__repr__()



from overrides import overrides
from typing import Optional

import torch

from allennlp.training.metrics.metric import Metric

from dygie.training.f1 import compute_f1

# TODO(dwadden) Need to use the decoded predictions so that we catch the gold examples longer than
# the span boundary.

class NERMetrics(Metric):
    """
    Computes precision, recall, and micro-averaged F1 from a list of predicted and gold labels.
    """
    def __init__(self, number_of_classes: int, none_label: int=0):
        self.number_of_classes = number_of_classes
        self.none_label = none_label
        self.reset()

    @overrides
    def __call__(self,
                 predictions: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: Optional[torch.Tensor] = None):
        predictions = predictions.cpu()
        gold_labels = gold_labels.cpu()
        mask = mask.cpu()
        for i in range(self.number_of_classes):
            if i == self.none_label:
                continue
            self._true_positives += ((predictions==i)*(gold_labels==i)*mask.bool()).sum().item()
            self._false_positives += ((predictions==i)*(gold_labels!=i)*mask.bool()).sum().item()
            self._true_negatives += ((predictions!=i)*(gold_labels!=i)*mask.bool()).sum().item()
            self._false_negatives += ((predictions!=i)*(gold_labels==i)*mask.bool()).sum().item()

    @overrides
    def get_metric(self, reset=False):
        """
        Returns
        -------
        A tuple of the following metrics based on the accumulated count statistics:
        precision : float
        recall : float
        f1-measure : float
        """
        predicted = self._true_positives + self._false_positives
        gold = self._true_positives + self._false_negatives
        matched = self._true_positives
        precision, recall, f1_measure = compute_f1(predicted, gold, matched)

        # Reset counts if at end of epoch.
        if reset:
            self.reset()

        return precision, recall, f1_measure

    @overrides
    def reset(self):
        self._true_positives = 0
        self._false_positives = 0
        self._true_negatives = 0
        self._false_negatives = 0



"""
Function to compute F1 scores.
"""


def safe_div(num, denom):
    if denom > 0:
        return num / denom
    else:
        return 0


def compute_f1(predicted, gold, matched):
    precision = safe_div(matched, predicted)
    recall = safe_div(matched, gold)
    f1 = safe_div(2 * precision * recall, precision + recall)
    return precision, recall, f1



# We have four types of outputs to leverage from radgraph
# 1. Entity token
# 2. Entity label
# 3. Relation tokens
# 4. Relation label

# Question for Jean Benoit: good idea to be based on f1 score ?
# We are using sets: so we do not account for repetitions of a word ! Except through relations ?

# We do three rewards for each type, so a total of six rewards
# +1 reward that does everything at the same time (maybe not that great)
# We combine these rewards with chosen coefficients (hyperparameters)

# To test our reward, we want that no token < token with wrong label < token with correct label
# Compares important clinical tokens of both texts
# Ex: is "pleural effusion" in both texts ?


# To be tested alone
def exact_entity_token_match_reward(hypothesis_annotation_list,
                                    reference_annotation_list):
    hypothesis_entity_token_list = set(
        map(lambda x: x['tokens'],
            hypothesis_annotation_list['entities'].values()))
    reference_entity_token_list = set(
        map(lambda x: x['tokens'],
            reference_annotation_list['entities'].values()))

    precision = sum([
        1 for x in hypothesis_entity_token_list
        if x in reference_entity_token_list
    ]) / len(hypothesis_entity_token_list)
    recall = sum([
        1 for x in reference_entity_token_list
        if x in hypothesis_entity_token_list
    ]) / len(reference_entity_token_list)
    f1_score = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

    return f1_score


# Compares important clinical entities and their syntaxtic use in both text
# Ex: is "pleural effusion" as a Definitively Absent Observation in both texts ?
# To be tested alone
def exact_entity_token_and_label_match_reward(hypothesis_annotation_list,
                                              reference_annotation_list):
    hypothesis_entity_token_and_label_list = set(
        map(lambda x: (x['tokens'], x['label']),
            hypothesis_annotation_list['entities'].values()))
    reference_entity_token_and_label_list = set(
        map(lambda x: (x['tokens'], x['label']),
            reference_annotation_list['entities'].values()))

    precision = sum([
        1 for x in hypothesis_entity_token_and_label_list
        if x in reference_entity_token_and_label_list
    ]) / len(hypothesis_entity_token_and_label_list)
    recall = sum([
        1 for x in reference_entity_token_and_label_list
        if x in hypothesis_entity_token_and_label_list
    ]) / len(reference_entity_token_and_label_list)
    f1_score = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

    return f1_score


# Ex: if "pleural effusion" is in both texts, is it as a Definitively Absent Observation in both texts?
# Not to be tested
def exact_entity_label_if_correct_token_match_reward(
        hypothesis_annotation_list, reference_annotation_list):
    hypothesis_entity_token_and_label_list = set(
        map(lambda x: (x['tokens'], x['label']),
            hypothesis_annotation_list['entities'].values()))
    reference_entity_token_and_label_list = set(
        map(lambda x: (x['tokens'], x['label']),
            reference_annotation_list['entities'].values()))

    hypothesis_entity_token_list = set(
        map(lambda x: x['tokens'],
            hypothesis_annotations['entities'].values()))
    reference_entity_token_list = set(
        map(lambda x: x['tokens'], reference_annotations['entities'].values()))

    hypothesis_entity_token_and_label_list = [
        (tokens, label)
        for (tokens, label) in hypothesis_entity_token_and_label_list
        if tokens in reference_entity_token_list
    ]
    reference_entity_token_and_label_list = [
        (tokens, label)
        for (tokens, label) in reference_entity_token_and_label_list
        if tokens in hypothesis_entity_token_list
    ]

    precision = sum([
        1 for x in hypothesis_entity_token_and_label_list
        if x in reference_entity_token_and_label_list
    ]) / len(hypothesis_entity_token_and_label_list) if len(hypothesis_entity_token_and_label_list) > 0 else 0.0

    recall = sum([
        1 for x in reference_entity_token_and_label_list
        if x in hypothesis_entity_token_and_label_list
    ]) / len(reference_entity_token_and_label_list) if len(reference_entity_token_and_label_list) > 0 else 0.0

    f1_score = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

    return f1_score


# Test alone, maybe best one ?
def partially_exact_entity_token_and_label_match_reward(
        hypothesis_annotation_list, reference_annotation_list):
    hypothesis_entity_token_and_label_list = set(
        map(lambda x: (x['tokens'], x['label']),
            hypothesis_annotation_list['entities'].values()))
    reference_entity_token_and_label_list = set(
        map(lambda x: (x['tokens'], x['label']),
            reference_annotation_list['entities'].values()))

    hypothesis_entity_token_list = set(
        map(lambda x: x['tokens'],
            hypothesis_annotation_list['entities'].values()))
    reference_entity_token_list = set(
        map(lambda x: x['tokens'],
            reference_annotation_list['entities'].values()))

    precision = sum([
                        1 for (x, y) in hypothesis_entity_token_and_label_list if
                        (x, y) in reference_entity_token_and_label_list
                    ] + [
                        0.5 for (x, y) in hypothesis_entity_token_and_label_list if (
                ((x, y) not in reference_entity_token_and_label_list) and
                (x in reference_entity_token_list))
                    ]) / len(hypothesis_entity_token_and_label_list) if len(
        hypothesis_entity_token_and_label_list) > 0 else 0.0

    recall = sum([
                     1 for (x, y) in reference_entity_token_and_label_list if
                     (x, y) in hypothesis_entity_token_and_label_list
                 ] + [
                     0.5 for (x, y) in reference_entity_token_and_label_list if (
                ((x, y) not in hypothesis_entity_token_and_label_list) and
                (x in hypothesis_entity_token_list))
                 ]) / len(reference_entity_token_and_label_list) if len(
        reference_entity_token_and_label_list) > 0 else 0.0

    f1_score = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

    return f1_score


# We do not do the partially exact loss for the relations, because does not make sense to do not have the label
# Just an exact match one that we combine with the other reward, and we are good

# Compares the relations, directly both the start and end tokens AS WELL AS the label
# Because a relation without the label does not really make sense


def exact_relation_token_and_label_match_reward(hypothesis_annotation_list,
                                                reference_annotation_list):
    hypothesis_relation_token_and_label_list = list(
        map(
            lambda x: [(x['tokens'], relation[0], relation[1])
                       for relation in x['relations']],
            hypothesis_annotation_list['entities'].values()))

    hypothesis_relation_token_and_label_list = set([
        (tokens_1, label,
         hypothesis_annotation_list['entities'][tokens_2]['tokens'])
        for relations in hypothesis_relation_token_and_label_list
        for (tokens_1, label, tokens_2) in relations
    ])

    reference_relation_token_and_label_list = list(
        map(
            lambda x: [(x['tokens'], relation[0], relation[1])
                       for relation in x['relations']],
            reference_annotation_list['entities'].values()))

    reference_relation_token_and_label_list = set([
        (tokens_1, label,
         reference_annotation_list['entities'][tokens_2]['tokens'])
        for relations in reference_relation_token_and_label_list
        for (tokens_1, label, tokens_2) in relations
    ])

    precision = sum([
        1 for x in hypothesis_relation_token_and_label_list
        if x in reference_relation_token_and_label_list
    ]) / len(hypothesis_relation_token_and_label_list) if len(hypothesis_relation_token_and_label_list) > 0 else 0.0

    recall = sum([
        1 for x in reference_relation_token_and_label_list
        if x in hypothesis_relation_token_and_label_list
    ]) / len(reference_relation_token_and_label_list) if len(reference_relation_token_and_label_list) > 0 else 0.0

    f1_score = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

    return f1_score


def partially_exact_relation_token_and_label_match_reward(
        hypothesis_annotation_list, reference_annotation_list):
    hypothesis_relation_token_and_label_list = list(
        map(
            lambda x: [(x['tokens'], relation[0], relation[1])
                       for relation in x['relations']],
            hypothesis_annotation_list['entities'].values()))

    hypothesis_relation_token_and_label_list = set([
        (tokens_1, label,
         hypothesis_annotation_list['entities'][tokens_2]['tokens'])
        for relations in hypothesis_relation_token_and_label_list
        for (tokens_1, label, tokens_2) in relations
    ])

    hypothesis_relation_token_list = set([
        (token_1, token_2)
        for (token_1, label,
             token_2) in hypothesis_relation_token_and_label_list
    ])

    reference_relation_token_and_label_list = list(
        map(
            lambda x: [(x['tokens'], relation[0], relation[1])
                       for relation in x['relations']],
            reference_annotation_list['entities'].values()))

    reference_relation_token_and_label_list = set([
        (tokens_1, label,
         reference_annotation_list['entities'][tokens_2]['tokens'])
        for relations in reference_relation_token_and_label_list
        for (tokens_1, label, tokens_2) in relations
    ])

    reference_relation_token_list = set([
        (token_1, token_2)
        for (token_1, label,
             token_2) in reference_relation_token_and_label_list
    ])

    precision = sum([
                        1 for (x, y, z) in hypothesis_relation_token_and_label_list if
                        (x, y, z) in reference_relation_token_and_label_list
                    ] + [
                        0.5 for (x, y, z) in hypothesis_relation_token_and_label_list if ((
                                                                                                  (x, y,
                                                                                                   z) not in reference_relation_token_and_label_list) and (
                                                                                                  (x,
                                                                                                   z) in reference_relation_token_list))
                    ]) / len(hypothesis_relation_token_and_label_list) if len(
        hypothesis_relation_token_and_label_list) > 0 else 0.0

    recall = sum([
                     1 for (x, y, z) in reference_relation_token_and_label_list if
                     (x, y, z) in hypothesis_relation_token_and_label_list
                 ] + [
                     0.5 for (x, y, z) in reference_relation_token_and_label_list if ((
                                                                                              (x, y,
                                                                                               z) not in hypothesis_relation_token_and_label_list) and (
                                                                                              (x,
                                                                                               z) in hypothesis_relation_token_list))
                 ]) / len(reference_relation_token_and_label_list) if len(
        reference_relation_token_and_label_list) > 0 else 0.0

    f1_score = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

    return f1_score


hypothesis_annotations = {
    'text':
        'FINAL REPORT INDICATION : ___ F with cough / / Cough TECHNIQUE : PA and lateral views of the chest . COMPARISON : None . FINDINGS : The lungs are clear without focal consolidation , , or edema . The cardiomediastinal silhouette is within normal limits . No acute osseous abnormalities . IMPRESSION : No acute cardiopulmonary process .',
    'entities': {
        '1': {
            'tokens': 'lungs',
            'label': 'ANAT-DP',
            'start_ix': 28,
            'end_ix': 28,
            'relations': []
        },
        '2': {
            'tokens': 'clear',
            'label': 'OBS-DP',
            'start_ix': 30,
            'end_ix': 30,
            'relations': [['located_at', '1']]
        },
        '3': {
            'tokens': 'focal',
            'label': 'OBS-DA',
            'start_ix': 32,
            'end_ix': 32,
            'relations': [['modify', '4']]
        },
        '4': {
            'tokens': 'consolidation',
            'label': 'OBS-DA',
            'start_ix': 33,
            'end_ix': 33,
            'relations': [['located_at', '1']]
        },
        '5': {
            'tokens': 'edema',
            'label': 'OBS-DA',
            'start_ix': 37,
            'end_ix': 37,
            'relations': []
        },
        '6': {
            'tokens': 'cardiomediastinal',
            'label': 'ANAT-DP',
            'start_ix': 40,
            'end_ix': 40,
            'relations': []
        },
        '7': {
            'tokens': 'silhouette',
            'label': 'ANAT-DP',
            'start_ix': 41,
            'end_ix': 41,
            'relations': [['modify', '6']]
        },
        '8': {
            'tokens': 'within',
            'label': 'OBS-DP',
            'start_ix': 43,
            'end_ix': 43,
            'relations': []
        },
        '9': {
            'tokens': 'normal',
            'label': 'OBS-DP',
            'start_ix': 44,
            'end_ix': 44,
            'relations': [['located_at', '6']]
        },
        '10': {
            'tokens': 'limits',
            'label': 'OBS-DP',
            'start_ix': 45,
            'end_ix': 45,
            'relations': [['modify', '9']]
        },
        '11': {
            'tokens': 'acute',
            'label': 'OBS-DA',
            'start_ix': 48,
            'end_ix': 48,
            'relations': [['modify', '13']]
        },
        '12': {
            'tokens': 'osseous',
            'label': 'ANAT-DP',
            'start_ix': 49,
            'end_ix': 49,
            'relations': []
        },
        '13': {
            'tokens': 'abnormalities',
            'label': 'OBS-DA',
            'start_ix': 50,
            'end_ix': 50,
            'relations': [['located_at', '12']]
        },
        '14': {
            'tokens': 'acute',
            'label': 'OBS-DA',
            'start_ix': 55,
            'end_ix': 55,
            'relations': [['modify', '16']]
        },
        '15': {
            'tokens': 'cardiopulmonary',
            'label': 'ANAT-DP',
            'start_ix': 56,
            'end_ix': 56,
            'relations': []
        },
        '16': {
            'tokens': 'process',
            'label': 'OBS-DA',
            'start_ix': 57,
            'end_ix': 57,
            'relations': [['located_at', '15']]
        }
    },
    'data_source': None,
    'data_split': 'inference'
}
reference_annotations = {
    'text':
        'FINAL REPORT INDICATION : ___ F with cough / / Cough TECHNIQUE : PA and lateral views of the chest . COMPARISON : None . FINDINGS : The lungs are clear without focal consolidation , , or edema . The cardiomediastinal silhouette is within normal limits . No acute osseous abnormalities . IMPRESSION : No acute cardiopulmonary process .',
    'entities': {
        '1': {
            'tokens': 'lungs',
            'label': 'ANAT-DP',
            'start_ix': 28,
            'end_ix': 28,
            'relations': []
        },
        '2': {
            'tokens': 'clear',
            'label': 'OBS-DP',
            'start_ix': 30,
            'end_ix': 30,
            'relations': [['located_at', '1']]
        },
        '3': {
            'tokens': 'focal',
            'label': 'OBS-DA',
            'start_ix': 32,
            'end_ix': 32,
            'relations': [['modify', '4']]
        },
        '4': {
            'tokens': 'consolidation',
            'label': 'OBS-DA',
            'start_ix': 33,
            'end_ix': 33,
            'relations': [['located_at', '1']]
        },
        '5': {
            'tokens': 'edema',
            'label': 'OBS-DA',
            'start_ix': 37,
            'end_ix': 37,
            'relations': []
        },
        '6': {
            'tokens': 'cardiomediastinal',
            'label': 'ANAT-DP',
            'start_ix': 40,
            'end_ix': 40,
            'relations': []
        },
        '7': {
            'tokens': 'silhouette',
            'label': 'ANAT-DP',
            'start_ix': 41,
            'end_ix': 41,
            'relations': [['modify', '6']]
        },
        '8': {
            'tokens': 'within',
            'label': 'OBS-DP',
            'start_ix': 43,
            'end_ix': 43,
            'relations': []
        },
        '9': {
            'tokens': 'normal',
            'label': 'OBS-DP',
            'start_ix': 44,
            'end_ix': 44,
            'relations': [['located_at', '6']]
        },
        '10': {
            'tokens': 'limits',
            'label': 'OBS-DP',
            'start_ix': 45,
            'end_ix': 45,
            'relations': [['modify', '9']]
        },
        '11': {
            'tokens': 'acute',
            'label': 'OBS-DA',
            'start_ix': 48,
            'end_ix': 48,
            'relations': [['modify', '13']]
        },
        '12': {
            'tokens': 'osseous',
            'label': 'ANAT-DP',
            'start_ix': 49,
            'end_ix': 49,
            'relations': []
        },
        '13': {
            'tokens': 'abnormalities',
            'label': 'OBS-DA',
            'start_ix': 50,
            'end_ix': 50,
            'relations': [['located_at', '12']]
        },
        '14': {
            'tokens': 'acute',
            'label': 'OBS-DA',
            'start_ix': 55,
            'end_ix': 55,
            'relations': [['modify', '16']]
        },
        '15': {
            'tokens': 'cardiopulmonary',
            'label': 'ANAT-DP',
            'start_ix': 56,
            'end_ix': 56,
            'relations': []
        },
        '16': {
            'tokens': 'process',
            'label': 'OBS-DA',
            'start_ix': 57,
            'end_ix': 57,
            'relations': [['located_at', '15']]
        }
    },
    'data_source': None,
    'data_split': 'inference'
}

# score = partially_exact_relation_token_and_label_match_reward(
#    hypothesis_annotations, reference_annotations)

# print(score)
# print(2 * (13 / 15) * (13 / 14) / (13 / 15 + 13 / 14))



import logging
from typing import Dict, List, Optional, Union
import copy

import torch
import torch.nn.functional as F
from overrides import overrides

from allennlp.data import Vocabulary
from allennlp.common.params import Params
from allennlp.models.model import Model
from allennlp.modules import TextFieldEmbedder, FeedForward, TimeDistributed
from allennlp.modules.span_extractors import EndpointSpanExtractor
from allennlp.nn import util, InitializerApplicator, RegularizerApplicator

# Import submodules.
from dygie.models.coref import CorefResolver
from dygie.models.ner import NERTagger
from dygie.models.relation import RelationExtractor
from dygie.models.events import EventExtractor
from dygie.data.dataset_readers import document

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@Model.register("dygie")
class DyGIE(Model):
    """
    TODO(dwadden) document me.

    Parameters
    ----------
    vocab : ``Vocabulary``
    text_field_embedder : ``TextFieldEmbedder``
        Used to embed the ``text`` ``TextField`` we get as input to the model.
    context_layer : ``Seq2SeqEncoder``
        This layer incorporates contextual information for each word in the document.
    feature_size: ``int``
        The embedding size for all the embedded features, such as distances or span widths.
    submodule_params: ``TODO(dwadden)``
        A nested dictionary specifying parameters to be passed on to initialize submodules.
    max_span_width: ``int``
        The maximum width of candidate spans.
    target_task: ``str``:
        The task used to make early stopping decisions.
    initializer : ``InitializerApplicator``, optional (default=``InitializerApplicator()``)
        Used to initialize the model parameters.
    module_initializer : ``InitializerApplicator``, optional (default=``InitializerApplicator()``)
        Used to initialize the individual modules.
    regularizer : ``RegularizerApplicator``, optional (default=``None``)
        If provided, will be used to calculate the regularization penalty during training.
    display_metrics: ``List[str]``. A list of the metrics that should be printed out during model
        training.
    """

    def __init__(self,
                 vocab: Vocabulary,
                 embedder: TextFieldEmbedder,
                 modules,  # TODO(dwadden) Add type.
                 feature_size: int,
                 max_span_width: int,
                 target_task: str,
                 feedforward_params: Dict[str, Union[int, float]],
                 loss_weights: Dict[str, float],
                 initializer: InitializerApplicator = InitializerApplicator(),
                 module_initializer: InitializerApplicator = InitializerApplicator(),
                 regularizer: Optional[RegularizerApplicator] = None,
                 display_metrics: List[str] = None) -> None:
        super(DyGIE, self).__init__(vocab, regularizer)

        ####################

        # Create span extractor.
        self._endpoint_span_extractor = EndpointSpanExtractor(
            embedder.get_output_dim(),
            combination="x,y",
            num_width_embeddings=max_span_width,
            span_width_embedding_dim=feature_size,
            bucket_widths=False)

        ####################

        # Set parameters.
        self._embedder = embedder
        self._loss_weights = loss_weights
        self._max_span_width = max_span_width
        self._display_metrics = self._get_display_metrics(target_task)
        token_emb_dim = self._embedder.get_output_dim()
        span_emb_dim = self._endpoint_span_extractor.get_output_dim()

        ####################

        # Create submodules.

        modules = Params(modules)

        # Helper function to create feedforward networks.
        def make_feedforward(input_dim):
            return FeedForward(input_dim=input_dim,
                               num_layers=feedforward_params["num_layers"],
                               hidden_dims=feedforward_params["hidden_dims"],
                               activations=torch.nn.ReLU(),
                               dropout=feedforward_params["dropout"])

        # Submodules

        self._ner = NERTagger.from_params(vocab=vocab,
                                          make_feedforward=make_feedforward,
                                          span_emb_dim=span_emb_dim,
                                          feature_size=feature_size,
                                          params=modules.pop("ner"))

        self._coref = CorefResolver.from_params(vocab=vocab,
                                                make_feedforward=make_feedforward,
                                                span_emb_dim=span_emb_dim,
                                                feature_size=feature_size,
                                                params=modules.pop("coref"))

        self._relation = RelationExtractor.from_params(vocab=vocab,
                                                       make_feedforward=make_feedforward,
                                                       span_emb_dim=span_emb_dim,
                                                       feature_size=feature_size,
                                                       params=modules.pop("relation"))

        self._events = EventExtractor.from_params(vocab=vocab,
                                                  make_feedforward=make_feedforward,
                                                  token_emb_dim=token_emb_dim,
                                                  span_emb_dim=span_emb_dim,
                                                  feature_size=feature_size,
                                                  params=modules.pop("events"))

        ####################

        # Initialize text embedder and all submodules
        for module in [self._ner, self._coref, self._relation, self._events]:
            module_initializer(module)

        initializer(self)

    @staticmethod
    def _get_display_metrics(target_task):
        """
        The `target` is the name of the task used to make early stopping decisions. Show metrics
        related to this task.
        """
        lookup = {
            "ner": [f"MEAN__{name}" for name in
                    ["ner_precision", "ner_recall", "ner_f1"]],
            "relation": [f"MEAN__{name}" for name in
                         ["relation_precision", "relation_recall", "relation_f1"]],
            "coref": ["coref_precision", "coref_recall", "coref_f1", "coref_mention_recall"],
            "events": [f"MEAN__{name}" for name in
                       ["trig_class_f1", "arg_class_f1"]]}
        if target_task not in lookup:
            raise ValueError(f"Invalied value {target_task} has been given as the target task.")
        return lookup[target_task]

    @staticmethod
    def _debatch(x):
        # TODO(dwadden) Get rid of this when I find a better way to do it.
        return x if x is None else x.squeeze(0)

    @overrides
    def forward(self,
                text,
                spans,
                metadata,
                ner_labels=None,
                coref_labels=None,
                relation_labels=None,
                trigger_labels=None,
                argument_labels=None):
        """
        TODO(dwadden) change this.
        """
        # In AllenNLP, AdjacencyFields are passed in as floats. This fixes it.
        if relation_labels is not None:
            relation_labels = relation_labels.long()
        if argument_labels is not None:
            argument_labels = argument_labels.long()

        # TODO(dwadden) Multi-document minibatching isn't supported yet. For now, get rid of the
        # extra dimension in the input tensors. Will return to this once the model runs.
        if len(metadata) > 1:
            raise NotImplementedError("Multi-document minibatching not yet supported.")

        metadata = metadata[0]
        spans = self._debatch(spans)  # (n_sents, max_n_spans, 2)
        ner_labels = self._debatch(ner_labels)  # (n_sents, max_n_spans)
        coref_labels = self._debatch(coref_labels)  #  (n_sents, max_n_spans)
        relation_labels = self._debatch(relation_labels)  # (n_sents, max_n_spans, max_n_spans)
        trigger_labels = self._debatch(trigger_labels)  # TODO(dwadden)
        argument_labels = self._debatch(argument_labels)  # TODO(dwadden)

        # Encode using BERT, then debatch.
        # Since the data are batched, we use `num_wrapping_dims=1` to unwrap the document dimension.
        # (1, n_sents, max_sententence_length, embedding_dim)

        # TODO(dwadden) Deal with the case where the input is longer than 512.
        text_embeddings = self._embedder(text, num_wrapping_dims=1)
        # (n_sents, max_n_wordpieces, embedding_dim)
        text_embeddings = self._debatch(text_embeddings)

        # (n_sents, max_sentence_length)
        text_mask = self._debatch(util.get_text_field_mask(text, num_wrapping_dims=1).float())
        sentence_lengths = text_mask.sum(dim=1).long()  # (n_sents)

        span_mask = (spans[:, :, 0] >= 0).float()  # (n_sents, max_n_spans)
        # SpanFields return -1 when they are used as padding. As we do some comparisons based on
        # span widths when we attend over the span representations that we generate from these
        # indices, we need them to be <= 0. This is only relevant in edge cases where the number of
        # spans we consider after the pruning stage is >= the total number of spans, because in this
        # case, it is possible we might consider a masked span.
        spans = F.relu(spans.float()).long()  # (n_sents, max_n_spans, 2)

        # Shape: (batch_size, num_spans, 2 * encoding_dim + feature_size)
        span_embeddings = self._endpoint_span_extractor(text_embeddings, spans)

        # Make calls out to the modules to get results.
        output_coref = {'loss': 0}
        output_ner = {'loss': 0}
        output_relation = {'loss': 0}
        output_events = {'loss': 0}

        # Prune and compute span representations for coreference module
        if self._loss_weights["coref"] > 0 or self._coref.coref_prop > 0:
            output_coref, coref_indices = self._coref.compute_representations(
                spans, span_mask, span_embeddings, sentence_lengths, coref_labels, metadata)

        # Propagation of global information to enhance the span embeddings
        if self._coref.coref_prop > 0:
            output_coref = self._coref.coref_propagation(output_coref)
            span_embeddings = self._coref.update_spans(
                output_coref, span_embeddings, coref_indices)

        # Make predictions and compute losses for each module
        if self._loss_weights['ner'] > 0:
            output_ner = self._ner(
                spans, span_mask, span_embeddings, sentence_lengths, ner_labels, metadata)

        if self._loss_weights['coref'] > 0:
            output_coref = self._coref.predict_labels(output_coref, metadata)

        if self._loss_weights['relation'] > 0:
            output_relation = self._relation(
                spans, span_mask, span_embeddings, sentence_lengths, relation_labels, metadata)

        if self._loss_weights['events'] > 0:
            # The `text_embeddings` serve as representations for event triggers.
            output_events = self._events(
                text_mask, text_embeddings, spans, span_mask, span_embeddings,
                sentence_lengths, trigger_labels, argument_labels,
                ner_labels, metadata)

        # Use `get` since there are some cases where the output dict won't have a loss - for
        # instance, when doing prediction.
        loss = (self._loss_weights['coref'] * output_coref.get("loss", 0) +
                self._loss_weights['ner'] * output_ner.get("loss", 0) +
                self._loss_weights['relation'] * output_relation.get("loss", 0) +
                self._loss_weights['events'] * output_events.get("loss", 0))

        # Multiply the loss by the weight multiplier for this document.
        weight = metadata.weight if metadata.weight is not None else 1.0
        loss *= torch.tensor(weight)

        output_dict = dict(coref=output_coref,
                           relation=output_relation,
                           ner=output_ner,
                           events=output_events)
        output_dict['loss'] = loss

        output_dict["metadata"] = metadata

        return output_dict

    def update_span_embeddings(self, span_embeddings, span_mask, top_span_embeddings,
                               top_span_mask, top_span_indices):
        # TODO(Ulme) Speed this up by tensorizing

        new_span_embeddings = span_embeddings.clone()
        for sample_nr in range(len(top_span_mask)):
            for top_span_nr, span_nr in enumerate(top_span_indices[sample_nr]):
                if top_span_mask[sample_nr, top_span_nr] == 0 or span_mask[sample_nr, span_nr] == 0:
                    break
                new_span_embeddings[sample_nr,
                                    span_nr] = top_span_embeddings[sample_nr, top_span_nr]
        return new_span_embeddings

    @overrides
    def make_output_human_readable(self, output_dict: Dict[str, torch.Tensor]):
        """
        Converts the list of spans and predicted antecedent indices into clusters
        of spans for each element in the batch.

        Parameters
        ----------
        output_dict : ``Dict[str, torch.Tensor]``, required.
            The result of calling :func:`forward` on an instance or batch of instances.

        Returns
        -------
        The same output dictionary, but with an additional ``clusters`` key:

        clusters : ``List[List[List[Tuple[int, int]]]]``
            A nested list, representing, for each instance in the batch, the list of clusters,
            which are in turn comprised of a list of (start, end) inclusive spans into the
            original document.
        """

        doc = copy.deepcopy(output_dict["metadata"])

        if self._loss_weights["coref"] > 0:
            # TODO(dwadden) Will need to get rid of the [0] when batch training is enabled.
            decoded_coref = self._coref.make_output_human_readable(output_dict["coref"])["predicted_clusters"][0]
            sentences = doc.sentences
            sentence_starts = [sent.sentence_start for sent in sentences]
            predicted_clusters = [document.Cluster(entry, i, sentences, sentence_starts)
                                  for i, entry in enumerate(decoded_coref)]
            doc.predicted_clusters = predicted_clusters
            # TODO(dwadden) update the sentences with cluster information.

        if self._loss_weights["ner"] > 0:
            for predictions, sentence in zip(output_dict["ner"]["predictions"], doc):
                sentence.predicted_ner = predictions

        if self._loss_weights["relation"] > 0:
            for predictions, sentence in zip(output_dict["relation"]["predictions"], doc):
                sentence.predicted_relations = predictions

        if self._loss_weights["events"] > 0:
            for predictions, sentence in zip(output_dict["events"]["predictions"], doc):
                sentence.predicted_events = predictions

        return doc

    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        """
        Get all metrics from all modules. For the ones that shouldn't be displayed, prefix their
        keys with an underscore.
        """
        metrics_coref = self._coref.get_metrics(reset=reset)
        metrics_ner = self._ner.get_metrics(reset=reset)
        metrics_relation = self._relation.get_metrics(reset=reset)
        metrics_events = self._events.get_metrics(reset=reset)

        # Make sure that there aren't any conflicting names.
        metric_names = (list(metrics_coref.keys()) + list(metrics_ner.keys()) +
                        list(metrics_relation.keys()) + list(metrics_events.keys()))
        assert len(set(metric_names)) == len(metric_names)
        all_metrics = dict(list(metrics_coref.items()) +
                           list(metrics_ner.items()) +
                           list(metrics_relation.items()) +
                           list(metrics_events.items()))

        # If no list of desired metrics given, display them all.
        if self._display_metrics is None:
            return all_metrics
        # Otherwise only display the selected ones.
        res = {}
        for k, v in all_metrics.items():
            if k in self._display_metrics:
                res[k] = v
            else:
                new_k = "_" + k
                res[new_k] = v
        return res



from overrides import overrides

from allennlp.training.metrics.metric import Metric

from dygie.training.f1 import compute_f1


class RelationMetrics(Metric):
    """
    Computes precision, recall, and micro-averaged F1 from a list of predicted and gold spans.
    """
    def __init__(self):
        self.reset()

    # TODO(dwadden) This requires decoding because the dataset reader gets rid of gold spans wider
    # than the span width. So, I can't just compare the tensor of gold labels to the tensor of
    # predicted labels.
    @overrides
    def __call__(self, predicted_relation_list, metadata_list):
        for predicted_relations, metadata in zip(predicted_relation_list, metadata_list):
            gold_relations = metadata.relation_dict
            self._total_gold += len(gold_relations)
            self._total_predicted += len(predicted_relations)
            for (span_1, span_2), label in predicted_relations.items():
                ix = (span_1, span_2)
                if ix in gold_relations and gold_relations[ix] == label:
                    self._total_matched += 1

    @overrides
    def get_metric(self, reset=False):
        precision, recall, f1 = compute_f1(self._total_predicted, self._total_gold, self._total_matched)

        # Reset counts if at end of epoch.
        if reset:
            self.reset()

        return precision, recall, f1

    @overrides
    def reset(self):
        self._total_gold = 0
        self._total_predicted = 0
        self._total_matched = 0



from typing import List
import numpy as np
import warnings

from overrides import overrides
import numpy
import json

from allennlp.common.util import JsonDict
from allennlp.nn import util
from allennlp.data import Batch
from allennlp.data import DatasetReader
from allennlp.models import Model
from allennlp.predictors.predictor import Predictor


@Predictor.register("dygie")
class DyGIEPredictor(Predictor):
    """
    Predictor for DyGIE model.

    If model was trained on coref, prediction is done on a whole document at
    once. This risks overflowing memory on large documents.
    If the model was trained without coref, prediction is done by sentence.
    """
    def __init__(
            self, model: Model, dataset_reader: DatasetReader) -> None:
        super().__init__(model, dataset_reader)

    def predict(self, document):
        return self.predict_json({"document": document})

    def predict_tokenized(self, tokenized_document: List[str]) -> JsonDict:
        instance = self._words_list_to_instance(tokenized_document)
        return self.predict_instance(instance)

    @overrides
    def dump_line(self, outputs):
        # Need to override to tell Python how to deal with Numpy ints.
        return json.dumps(outputs, default=int) + "\n"

    # TODO(dwadden) Can this be implemented in `forward_on_instance`  instead?
    @overrides
    def predict_instance(self, instance):
        """
        An instance is an entire document, represented as a list of sentences.
        """
        model = self._model
        cuda_device = model._get_prediction_device()

        # Try to predict this batch.
        try:
            dataset = Batch([instance])
            dataset.index_instances(model.vocab)
            model_input = util.move_to_device(dataset.as_tensor_dict(), cuda_device)
            prediction = model.make_output_human_readable(model(**model_input)).to_json()
        # If we run out of GPU memory, warn user and indicate that this document failed.
        # This way, prediction doesn't grind to a halt every time we run out of GPU.
        except RuntimeError as err:
            # doc_key, dataset, sentences, message
            metadata = instance["metadata"].metadata
            doc_key = metadata.doc_key
            msg = (f"Encountered a RunTimeError on document {doc_key}. Skipping this example."
                   f" Error message:\n{err.args[0]}.")
            warnings.warn(msg)
            prediction = metadata.to_json()
            prediction["_FAILED_PREDICTION"] = True

        return prediction



"""
This is basically a copy of AllenNLP's Pruner module, but with support for entity beams.
"""

from typing import Tuple, Union

from overrides import overrides
import torch

from allennlp.nn import util
from allennlp.modules import TimeDistributed


def make_pruner(scorer, entity_beam=False, gold_beam=False):
    """
    Create a pruner that either takes outputs of other scorers (i.e. entity beam), or uses its own
    scorer (the `default_scorer`).
    """
    item_scorer = torch.nn.Sequential(
        TimeDistributed(scorer),
        TimeDistributed(torch.nn.Linear(scorer.get_output_dim(), 1)))
    min_score_to_keep = 1e-10 if entity_beam else None

    return Pruner(item_scorer, entity_beam, gold_beam, min_score_to_keep)


class Pruner(torch.nn.Module):
    """
    This module scores and prunes items in a list using a parameterised scoring function and a
    threshold.

    Parameters
    ----------
    scorer : ``torch.nn.Module``, required.
        A module which, given a tensor of shape (batch_size, num_items, embedding_size),
        produces a tensor of shape (batch_size, num_items, 1), representing a scalar score
        per item in the tensor.
    entity_beam: bool, optional.
        If True, use class scores output from another module instead of using own scorer.
    gold_beam: bool, optional.
       If True, use gold arguments.
    min_score_to_keep : float, optional.
        If given, only keep items that score at least this high.
    """
    def __init__(self, scorer: torch.nn.Module, entity_beam: bool = False, gold_beam: bool = False,
                 min_score_to_keep: float = None) -> None:
        super().__init__()
        # If gold beam is on, then entity beam must be off and min_score_to_keep must be None.
        assert not (gold_beam and ((min_score_to_keep is not None) or entity_beam))
        self._scorer = scorer
        self._entity_beam = entity_beam
        self._gold_beam = gold_beam
        self._min_score_to_keep = min_score_to_keep

    @overrides
    def forward(self, # pylint: disable=arguments-differ
                embeddings: torch.FloatTensor,
                mask: torch.LongTensor,
                num_items_to_keep: Union[int, torch.LongTensor],
                class_scores: torch.FloatTensor = None,
                gold_labels: torch.long = None) -> Tuple[torch.FloatTensor, torch.LongTensor,
                                                         torch.LongTensor, torch.FloatTensor]:
        """
        Extracts the top-k scoring items with respect to the scorer. We additionally return
        the indices of the top-k in their original order, not ordered by score, so that downstream
        components can rely on the original ordering (e.g., for knowing what spans are valid
        antecedents in a coreference resolution model). May use the same k for all sentences in
        minibatch, or different k for each.

        Parameters
        ----------
        embeddings : ``torch.FloatTensor``, required.
            A tensor of shape (batch_size, num_items, embedding_size), containing an embedding for
            each item in the list that we want to prune.
        mask : ``torch.LongTensor``, required.
            A tensor of shape (batch_size, num_items), denoting unpadded elements of
            ``embeddings``.
        num_items_to_keep : ``Union[int, torch.LongTensor]``, required.
            If a tensor of shape (batch_size), specifies the number of items to keep for each
            individual sentence in minibatch.
            If an int, keep the same number of items for all sentences.
        class_scores:
           Class scores to be used with entity beam.
        candidate_labels: If in debugging mode, use gold labels to get beam.

        Returns
        -------
        top_embeddings : ``torch.FloatTensor``
            The representations of the top-k scoring items.
            Has shape (batch_size, max_num_items_to_keep, embedding_size).
        top_mask : ``torch.LongTensor``
            The corresponding mask for ``top_embeddings``.
            Has shape (batch_size, max_num_items_to_keep).
        top_indices : ``torch.IntTensor``
            The indices of the top-k scoring items into the original ``embeddings``
            tensor. This is returned because it can be useful to retain pointers to
            the original items, if each item is being scored by multiple distinct
            scorers, for instance. Has shape (batch_size, max_num_items_to_keep).
        top_item_scores : ``torch.FloatTensor``
            The values of the top-k scoring items.
            Has shape (batch_size, max_num_items_to_keep, 1).
        num_items_kept
        """
        # If an int was given for number of items to keep, construct tensor by repeating the value.
        if isinstance(num_items_to_keep, int):
            batch_size = mask.size(0)
            # Put the tensor on same device as the mask.
            num_items_to_keep = num_items_to_keep * torch.ones([batch_size], dtype=torch.long,
                                                               device=mask.device)

        mask = mask.unsqueeze(-1)
        num_items = embeddings.size(1)

        # Shape: (batch_size, num_items, 1)
        # If entity beam is one, use the class scores. Else ignore them and use the scorer.
        if self._entity_beam:
            scores, _ = class_scores.max(dim=-1)
            scores = scores.unsqueeze(-1)
        # If gold beam is one, give a score of 0 wherever the gold label is non-zero (indicating a
        # non-null label), otherwise give a large negative number.
        elif self._gold_beam:
            scores = torch.where(gold_labels > 0,
                                 torch.zeros_like(gold_labels, dtype=torch.float),
                                 -1e20 * torch.ones_like(gold_labels, dtype=torch.float))
            scores = scores.unsqueeze(-1)
        else:
            scores = self._scorer(embeddings)

        # If we're only keeping items that score above a given threshold, change the number of kept
        # items here.
        if self._min_score_to_keep is not None:
            num_good_items = torch.sum(scores > self._min_score_to_keep, dim=1).squeeze()
            num_items_to_keep = torch.min(num_items_to_keep, num_good_items)
        # If gold beam is on, keep the gold items.
        if self._gold_beam:
            num_items_to_keep = torch.sum(gold_labels > 0, dim=1)

        # Always keep at least one item to avoid edge case with empty matrix.
        max_items_to_keep = max(num_items_to_keep.max().item(), 1)

        if scores.size(-1) != 1 or scores.dim() != 3:
            raise ValueError(f"The scorer passed to Pruner must produce a tensor of shape"
                             f"(batch_size, num_items, 1), but found shape {scores.size()}")
        # Make sure that we don't select any masked items by setting their scores to be very
        # negative.  These are logits, typically, so -1e20 should be plenty negative.
        # NOTE(`mask` needs to be a byte tensor now.)
        scores = util.replace_masked_values(scores, mask.bool(), -1e20)

        # Shape: (batch_size, max_num_items_to_keep, 1)
        _, top_indices = scores.topk(max_items_to_keep, 1)

        # Mask based on number of items to keep for each sentence.
        # Shape: (batch_size, max_num_items_to_keep)
        top_indices_mask = util.get_mask_from_sequence_lengths(num_items_to_keep, max_items_to_keep)
        top_indices_mask = top_indices_mask.bool()

        # Shape: (batch_size, max_num_items_to_keep)
        top_indices = top_indices.squeeze(-1)

        # Fill all masked indices with largest "top" index for that sentence, so that all masked
        # indices will be sorted to the end.
        # Shape: (batch_size, 1)
        fill_value, _ = top_indices.max(dim=1)
        fill_value = fill_value.unsqueeze(-1)
        # Shape: (batch_size, max_num_items_to_keep)
        top_indices = torch.where(top_indices_mask, top_indices, fill_value)

        # Now we order the selected indices in increasing order with
        # respect to their indices (and hence, with respect to the
        # order they originally appeared in the ``embeddings`` tensor).
        top_indices, _ = torch.sort(top_indices, 1)

        # Shape: (batch_size * max_num_items_to_keep)
        # torch.index_select only accepts 1D indices, but here
        # we need to select items for each element in the batch.
        flat_top_indices = util.flatten_and_batch_shift_indices(top_indices, num_items)

        # Shape: (batch_size, max_num_items_to_keep, embedding_size)
        top_embeddings = util.batched_index_select(embeddings, top_indices, flat_top_indices)

        # Combine the masks on spans that are out-of-bounds, and the mask on spans that are outside
        # the top k for each sentence.
        # Shape: (batch_size, max_num_items_to_keep)
        sequence_mask = util.batched_index_select(mask, top_indices, flat_top_indices)
        sequence_mask = sequence_mask.squeeze(-1).bool()
        top_mask = top_indices_mask & sequence_mask
        top_mask = top_mask.long()

        # Shape: (batch_size, max_num_items_to_keep, 1)
        top_scores = util.batched_index_select(scores, top_indices, flat_top_indices)

        return top_embeddings, top_mask, top_indices, top_scores, num_items_to_keep



import json
import re
from .reward_functions import exact_entity_token_match_reward, partially_exact_relation_token_and_label_match_reward, \
    partially_exact_entity_token_and_label_match_reward, \
    exact_entity_token_and_label_match_reward, exact_relation_token_and_label_match_reward


def compute_reward(hypothesis_annotation_list, reference_annotation_list,
                   lambda_e, lambda_r, reward_level):
    if len(hypothesis_annotation_list["entities"].keys()) == 0 or len(
            reference_annotation_list["entities"].keys()) == 0:
        return 0

    if reward_level == "simple":
        return lambda_e * exact_entity_token_match_reward(
            hypothesis_annotation_list, reference_annotation_list)
    elif reward_level == "partial":
        return lambda_e * partially_exact_entity_token_and_label_match_reward(
            hypothesis_annotation_list, reference_annotation_list
        ) + lambda_r * partially_exact_relation_token_and_label_match_reward(
            hypothesis_annotation_list, reference_annotation_list)
    elif reward_level == "complete":
        return lambda_e * exact_entity_token_and_label_match_reward(
            hypothesis_annotation_list, reference_annotation_list
        ) + lambda_r * exact_relation_token_and_label_match_reward(
            hypothesis_annotation_list, reference_annotation_list)


def preprocess_reports(report_list):
    """ Load up the files mentioned in the temporary json file, and
    processes them in format that the dygie model can take as input.
    Also save the processed file in a temporary file.
    """
    final_list = []
    for idx, report in enumerate(report_list):
        sen = re.sub('(?<! )(?=[/,-,:,.,!?()])|(?<=[/,-,:,.,!?()])(?! )', r' ',
                     report).split()
        temp_dict = {}

        temp_dict["doc_key"] = str(idx)

        ## Current way of inference takes in the whole report as 1 sentence
        temp_dict["sentences"] = [sen]

        final_list.append(temp_dict)

    return [json.dumps(item) for item in final_list]


def postprocess_reports(results):
    """Post processes all the reports and saves the result in train.json format
    """
    final_dict = {}
    data = []

    for r in results:
        data.append(json.loads(r))

    for file in data:
        postprocess_individual_report(file, final_dict)

    return final_dict


def postprocess_individual_report(file, final_dict, data_source=None):
    """Postprocesses individual report

    Args:
        file: output dict for individual reports
        final_dict: Dict for storing all the reports
    """

    try:
        temp_dict = {}

        temp_dict['text'] = " ".join(file['sentences'][0])
        n = file['predicted_ner'][0]
        r = file['predicted_relations'][0]
        s = file['sentences'][0]
        temp_dict["entities"] = get_entity(n, r, s)
        temp_dict["data_source"] = data_source
        temp_dict["data_split"] = "inference"

        final_dict[file['doc_key']] = temp_dict

    except:
        print(
            f"Error in doc key: {file['doc_key']}. Skipping inference on this file"
        )


def get_entity(n, r, s):
    """Gets the entities for individual reports

    Args:
        n: list of entities in the report
        r: list of relations in the report
        s: list containing tokens of the sentence

    Returns:
        dict_entity: Dictionary containing the entites in the format similar to train.json

    """

    dict_entity = {}
    rel_list = [item[0:2] for item in r]
    ner_list = [item[0:2] for item in n]
    for idx, item in enumerate(n):
        temp_dict = {}
        start_idx, end_idx, label = item[0], item[1], item[2]
        temp_dict['tokens'] = " ".join(s[start_idx:end_idx + 1])
        temp_dict['label'] = label
        temp_dict['start_ix'] = start_idx
        temp_dict['end_ix'] = end_idx
        rel = []
        relation_idx = [
            i for i, val in enumerate(rel_list) if val == [start_idx, end_idx]
        ]
        for i, val in enumerate(relation_idx):
            obj = r[val][2:4]
            lab = r[val][4]
            try:
                object_idx = ner_list.index(obj) + 1
            except:
                continue
            rel.append([lab, str(object_idx)])
        temp_dict['relations'] = rel
        dict_entity[str(idx + 1)] = temp_dict

    return dict_entity
